import mock
import pytest

from swimlane.core.fields.reference import ReferenceCursor
from swimlane.core.resources.record import Record
from swimlane.core.resources.app import App
from swimlane.exceptions import ValidationError, SwimlaneHTTP400Error


class TestReferenceField(object):

    multi_field_name = 'Reference'
    single_field_name = 'Reference (Select Single)'

    def test_target_app(self, mock_record, mock_swimlane, mock_app):
        """Test availability of .target_app property on ReferenceCursor"""

        field = mock_record.get_field(self.multi_field_name)

        with mock.patch.object(mock_swimlane.apps, 'get', return_value=mock_app):

            assert field.target_app == mock_app

            reference_cursor = mock_record[self.multi_field_name]
            assert isinstance(reference_cursor, ReferenceCursor)
            assert reference_cursor.target_app is field.target_app

    def test_single_select(self, mock_swimlane, mock_record):
        """Test single-select get/set"""

        with mock.patch.object(mock_swimlane.apps, 'get', return_value=mock_record.app):
            field = mock_record.get_field(self.single_field_name)

            assert field.get_python() is None

            # Assert setting to None adds the field's key to raw values
            field.set_python(None)
            assert mock_record._raw['values'][field.id] is None

            field.set_python(mock_record)
            assert mock_record._raw['values'][field.id] == mock_record.id

            field.set_swimlane(mock_record.id)
            assert mock_record.id in field._value

    def test_empty_multi_select(self, mock_record):
        """Test empty multi select reference field"""

        field = mock_record.get_field('Reference')
        mock_record['Reference'] = None
        swimlane = field.get_swimlane()
        python = field.get_python()
        assert swimlane is None
        assert len(python) == 0

    def test_lazy_retrieval(self, mock_app, mock_record):
        """Test lazy retrieval of referenced records"""
        # Can't patch attribute
        field = mock_record.get_field(self.multi_field_name)
        field._ReferenceField__target_app = mock_app
        reference_cursor = mock_record[self.multi_field_name]

        with mock.patch.object(reference_cursor.target_app.records, 'get') as mock_record_get:
            mock_record_get.return_value = mock_record
            assert mock_record_get.call_count == 0

            # Lazy retrieval of target app definition and selected records

            assert len(reference_cursor) == 3
            assert mock_record_get.call_count == 3

            for referenced_record in reference_cursor:
                assert isinstance(referenced_record, Record)

            # Add/remove/set referenced record(s) by Record instance or ID
            mock_record[self.multi_field_name] = [mock_record]

            # Select/deselect
            assert len(mock_record[self.multi_field_name]) == 1
            mock_record[self.multi_field_name].remove(mock_record)
            assert len(mock_record[self.multi_field_name]) == 0
            mock_record[self.multi_field_name].add(mock_record)
            assert len(mock_record[self.multi_field_name]) == 1

            # No new requests should have been necessary, other than the original lookups
            assert mock_record_get.call_count == 3

    def test_target_app_validation(self, mock_swimlane, mock_record, mock_app):
        """Test that reference field validation checks that provided record is in the field's target app"""
        # Set target app to some random other app
        target_app = App(mock_swimlane, mock_app._raw)
        target_app.id = 'abcdef'
        target_app.name = 'Some other app'

        with mock.patch.object(mock_swimlane.apps, 'get', return_value=target_app):
            reference_cursor = mock_record[self.multi_field_name]

            with pytest.raises(ValidationError):
                reference_cursor.add(mock_record)

            with pytest.raises(ValidationError):
                mock_record[self.multi_field_name] = [mock_record]

    def test_orphaned_record_cleanup(self, mock_swimlane, mock_record, mock_app):
        """Test that orphaned records that no longer exist are ignored by ReferenceCursor"""
        mock_response = mock.MagicMock()
        mock_response.status_code = 400
        mock_response.json.return_value = {
            'ErrorCode': 3002,
            'Argument': None
        }

        with mock.patch.object(mock_swimlane.apps, 'get', return_value=mock_app):
            with mock.patch.object(mock_app.records, 'get', side_effect=SwimlaneHTTP400Error(mock_response)):
                assert len(mock_record[self.multi_field_name]) == 0

    def test_get_report(self, mock_record):
        """All for total coverage"""
        assert mock_record.get_field(self.single_field_name).get_report(mock_record) == mock_record.id



"""Module providing support for automatic APIResource caching

A ResourcesCache instance is provided on all Swimlane client instances automatically

.. versionadded:: 2.16.2
"""
import copy
import functools
import logging
from collections import defaultdict

from cachetools import LFUCache

from swimlane.core.resources.base import APIResource


logger = logging.getLogger(__name__)


class ResourcesCache(object):
    """Universal APIResource instance cache

    Uses separate caches per APIResource type, and provides mapping between available cache keys and real cache
    primary key automatically
    """

    def __init__(self, per_cache_max_size):
        self.__cache_max_size = per_cache_max_size
        self.__caches = defaultdict(self.__cache_factory)
        self.__cache_index_key_map = {}

        if self.__cache_max_size == 0:
            logger.warning('Cache size set to 0, resource caching disabled')

    def __len__(self):
        """Return sum of all cache sizes"""
        return sum(c.currsize for c in self.__caches.values())

    def __contains__(self, item):
        """Check if resource is in cache, expects same 3-length tuple key as __getitem__"""
        index_key = get_cache_index_key(item)
        cache_key = self.__cache_index_key_map.get(index_key)
        target_cache = self.__caches[index_key[0]]
        return cache_key in target_cache

    def __getitem__(self, item):
        """Get cached resource, expects item to be 3-length tuple of (resource class, target key, target value)"""
        key = get_cache_index_key(item)
        cls = key[0]

        # Check if in any fields index
        cache_internal_key = self.__cache_index_key_map[key]

        try:
            # Return copy of cached object
            return copy.copy(self.__caches[cls][cache_internal_key])
        except KeyError:
            # Internal cache miss for target resource, quietly remove from cache key map and let error bubble
            self.__cache_index_key_map.pop(key, None)
            raise

    def __delitem__(self, resource):
        """Remove resource instance from internal cache"""
        self.__caches[type(resource)].pop(resource.get_cache_internal_key(), None)

    def __cache_factory(self):
        """Build and return a new cache instance"""
        return LFUCache(self.__cache_max_size)

    def cache(self, resource):
        """Insert a resource instance into appropriate resource cache"""
        if not isinstance(resource, APIResource):
            raise TypeError('Cannot cache `{!r}`, can only cache APIResource instances'.format(resource))

        # Disable inserts to cache when disabled
        if self.__cache_max_size == 0:
            return

        try:
            cache_internal_key = resource.get_cache_internal_key()
            cache_index_keys = resource.get_cache_index_keys().items()
        except NotImplementedError:
            logger.warning(
                'Not caching `{!r}`, resource did not provide all necessary cache details'.format(resource)
            )
        else:
            resource_type = type(resource)

            for key, value in cache_index_keys:
                self.__cache_index_key_map[(resource_type, key, value)] = cache_internal_key

            self.__caches[resource_type][cache_internal_key] = resource

            logger.debug('Cached `{!r}`'.format(resource))

    def clear(self, *resource_types):
        """Clear cache for each provided APIResource class, or all resources if no classes are provided"""
        resource_types = resource_types or tuple(self.__caches.keys())

        for cls in resource_types:
            # Clear and delete cache instances to guarantee no lingering references
            self.__caches[cls].clear()
            del self.__caches[cls]


def get_cache_index_key(resource):
    """Return a usable cache lookup key for an already initialized resource

    Args:
        resource (APIResource|tuple): APIResource instance or 3-length tuple key returned from this function

    Raises:
        TypeError: If resource is not an APIResource instance or acceptable 3-length tuple cache key
    """
    if isinstance(resource, APIResource):
        attr, attr_value = list(resource.get_cache_index_keys().items())[0]
        key = (type(resource), attr, attr_value)
    else:
        key = tuple(resource)

    if len(key) != 3:
        raise TypeError('Cache key must be tuple of (class, key, value), got `{!r}` instead'.format(key))

    if not issubclass(key[0], APIResource):
        raise TypeError('First value of cache key must be a subclass of APIResource, got `{!r}` instead'.format(key[0]))

    return key


def check_cache(resource_type):
    """Decorator for adapter methods to check cache for resource before normally sending requests to retrieve data

    Only works with single kwargs, almost always used with @one_of_keyword_only decorator

    Args:
        resource_type (type(APIResource)): Subclass of APIResource of cache to be checked when called
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                adapter = args[0]
                key, val = list(kwargs.items())[0]
            except IndexError:
                logger.warning("Couldn't generate full index key, skipping cache")
            else:

                index_key = (resource_type, key, val)
                try:
                    cached_record = adapter._swimlane.resources_cache[index_key]
                except KeyError:
                    logger.debug('Cache miss: `{!r}`'.format(index_key))
                else:
                    logger.debug('Cache hit: `{!r}`'.format(cached_record))
                    return cached_record

            # Fallback to default function call
            return func(*args, **kwargs)

        return wrapper
    return decorator



import mock

from swimlane.core.resources.report import Report

raw_report = {'$type': 'Core.Models.Search.Report, Core',
              'aggregates': [],
              'allowed': [],
              'applicationIds': ['58de205f07637a0264c0ccbf'],
              'columns': ['58de205f07637a0264c0ccc1'],
              'createdByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                'id': '58de1d1c07637a0264c0ca6a',
                                'name': 'admin'},
              'createdDate': '2017-03-31T09:24:47.758Z',
              'defaultSearchReport': True,
              'disabled': False,
              'filters': [],
              'groupBys': [],
              'id': '58de205f07637a0264c0ccc5',
              'modifiedByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                 'id': '58de1d1c07637a0264c0ca6a',
                                 'name': 'admin'},
              'modifiedDate': '2017-03-31T09:24:47.758Z',
              'name': 'Default',
              'offset': 0,
              'pageSize': 10,
              'permissions': {'$type': 'Core.Models.Security.PermissionMatrix, Core'},
              'sorts': {
                  '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[Core.Models.Search.SortTypes, Core]], mscorlib',
                  '58de205f07637a0264c0ccc1': 'ascending'}}


def test_list(mock_app, mock_swimlane):
    mock_response = mock.MagicMock()
    mock_response.json.return_value = [raw_report for _ in range(3)]

    with mock.patch.object(mock_swimlane, 'request', return_value=mock_response):
        reports = mock_app.reports.list()
        assert len(reports) == 3
        for report in reports:
            assert isinstance(report, Report)


def test_get(mock_app, mock_swimlane):
    mock_response = mock.MagicMock()
    mock_response.json.return_value = raw_report

    with mock.patch.object(mock_swimlane, 'request', return_value=mock_response):
        report = mock_app.reports.get(report_id='report_id')
        assert isinstance(report, Report)
